import static org.jocl.CL.*;

import org.jocl.*;
import java.util.Arrays;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;


public class JoclOuter
{
    private static String programSource =
        "__kernel void "+
        "sampleKernel(__global const int *a,"+
        "             __global const int *b,"+
        "             __global int *c)"+
        "{"+
        "    int gid = get_global_id(0);"+
        "    c[gid] = a[gid] * b[gid];"+
        "}";

    public static int[] computeRow(int[] a, int[] b)
    {
        // Create input- and output data
        int n = a.length;
        int srcArrayA[] = a;
        int srcArrayB[] = b;
        int dstArray[] = new int[n];

        Pointer srcA = Pointer.to(srcArrayA);
        Pointer srcB = Pointer.to(srcArrayB);
        Pointer dst = Pointer.to(dstArray);

        // The platform, device type and device number
        // that will be used
        final int platformIndex = 0;
        final long deviceType = CL_DEVICE_TYPE_ALL;
        final int deviceIndex = 0;

        // Enable exceptions and subsequently omit error checks in this sample
        CL.setExceptionsEnabled(true);

        // Obtain the number of platforms
        int numPlatformsArray[] = new int[1];
        clGetPlatformIDs(0, null, numPlatformsArray);
        int numPlatforms = numPlatformsArray[0];

        // Obtain a platform ID
        cl_platform_id platforms[] = new cl_platform_id[numPlatforms];
        clGetPlatformIDs(platforms.length, platforms, null);
        cl_platform_id platform = platforms[platformIndex];

        // Initialize the context properties
        cl_context_properties contextProperties = new cl_context_properties();
        contextProperties.addProperty(CL_CONTEXT_PLATFORM, platform);

        // Obtain the number of devices for the platform
        int numDevicesArray[] = new int[1];
        clGetDeviceIDs(platform, deviceType, 0, null, numDevicesArray);
        int numDevices = numDevicesArray[0];

        // Obtain a device ID
        cl_device_id devices[] = new cl_device_id[numDevices];
        clGetDeviceIDs(platform, deviceType, numDevices, devices, null);
        cl_device_id device = devices[deviceIndex];

        // Create a context for the selected device
        cl_context context = clCreateContext(
            contextProperties, 1, new cl_device_id[]{device},
            null, null, null);

        cl_command_queue commandQueue =
            clCreateCommandQueue(context, device, 0, null);

        // Allocate the memory objects for the input- and output data
        cl_mem memObjects[] = new cl_mem[3];
        memObjects[0] = clCreateBuffer(context,
            CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
            Sizeof.cl_int * n, srcA, null);
        memObjects[1] = clCreateBuffer(context,
            CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
            Sizeof.cl_int * n, srcB, null);
        memObjects[2] = clCreateBuffer(context,
            CL_MEM_READ_WRITE,
            Sizeof.cl_int * n, null, null);

        // Create the program from the source code
        cl_program program = clCreateProgramWithSource(context,
            1, new String[]{ programSource }, null, null);

        // Build the program
        clBuildProgram(program, 0, null, null, null, null);

        // Create the kernel
        cl_kernel kernel = clCreateKernel(program, "sampleKernel", null);

        // Set the arguments for the kernel
        clSetKernelArg(kernel, 0,
            Sizeof.cl_mem, Pointer.to(memObjects[0]));
        clSetKernelArg(kernel, 1,
            Sizeof.cl_mem, Pointer.to(memObjects[1]));
        clSetKernelArg(kernel, 2,
            Sizeof.cl_mem, Pointer.to(memObjects[2]));

        // Set the work-item dimensions
        long global_work_size[] = new long[]{n};
        long local_work_size[] = new long[]{1};

        // Execute the kernel
        clEnqueueNDRangeKernel(commandQueue, kernel, 1, null,
            global_work_size, local_work_size, 0, null, null);

        // Read the output data
        clEnqueueReadBuffer(commandQueue, memObjects[2], CL_TRUE, 0,
            n * Sizeof.cl_int, dst, 0, null, null);

        // Release kernel, program, and memory objects
        clReleaseMemObject(memObjects[0]);
        clReleaseMemObject(memObjects[1]);
        clReleaseMemObject(memObjects[2]);
        clReleaseKernel(kernel);
        clReleaseProgram(program);
        clReleaseCommandQueue(commandQueue);
        clReleaseContext(context);

        return dstArray;
    }

    public static void main (String[] args) {

        // Reading in data
        String vs = "";
        String ws = "";
        try {
          BufferedReader in = new BufferedReader(new FileReader(args[0]));
          String s = in.readLine();
          String[] nums;

          do {
            nums = s.split(",");
            vs += nums[0]+",";
            ws += nums[1]+",";
          } while ((s = in.readLine()) != null);

          in.close();
        } catch (IOException e) {
          System.out.println("some error");
        }

        // create vectors
        int[] v = Arrays.stream(vs.split(",")).mapToInt(Integer::parseInt).toArray();
        int[] w = Arrays.stream(ws.split(",")).mapToInt(Integer::parseInt).toArray();

        int dim = v.length;
        int[][] result = new int[dim][dim];

        // Multiple calls to GPU here
        // long startTime = System.nanoTime();
        long startTime = 0;
        long endTime = 0;
        for (int i = 0; i < dim; i++) {
          int[] a = new int[dim];
          Arrays.fill(a, v[i]);
          startTime = System.nanoTime();
          result[i] = computeRow(a, w);
          endTime = System.nanoTime();
        }
        // long endTime = System.nanoTime();

        for (int i = 0; i < result.length; i++) {
          System.out.print(result[i][0]);
          for (int j = 1; j < result.length; j++) {
            System.out.print(","+result[i][j]);
          }
          System.out.println();
        }

        // put any argument after data to display execution time
        if (args.length > 1) {
          try
          {
              Thread.sleep(1800);
          }
          catch(InterruptedException ex)
          {
              Thread.currentThread().interrupt();
          }
          System.out.println();
          System.out.print("Total Execution Time: " + (endTime - startTime));
          System.out.println(" [ns]");
        }
    }
}
